/*
 * A Functional Cardboard Record Player Based on an ESP32.
 * By CREATIVE LABS
*/
#include <SPI.h>
#include <MFRC522.h>
#include <Stepper.h>
#include <DFRobotDFPlayerMini.h>

// RFID setup
#define RST_PIN 22
#define SS_PIN 21
MFRC522 rfid(SS_PIN, RST_PIN);

// Stepper motor setup
#define STEPS 2048
Stepper stepper(STEPS, 32, 33, 14, 12);

// DFPlayer Mini setup
#define DFPLAYER_RX 16
#define DFPLAYER_TX 17
HardwareSerial mySerial(1);

DFRobotDFPlayerMini myDFPlayer;

// Variables to manage song playing and RFID handling
unsigned long lastRFIDTime = 0;
bool songPlaying = false;
byte lastUID[10];
byte lastUIDSize = 0;

// Define UIDs of the NFC chips. Each UID is represented in hexadecimal format.
// For example: 04:46:B3:CA:78:00:00 becomes {0x04, 0x46, 0xB3, 0xCA, 0x78, 0x00, 0x00}
byte targetUID1[] = {0x04, 0xAE, 0xF2, 0x97, 0x78, 0x00, 0x00};
byte targetUID2[] = {0x04, 0x46, 0xB3, 0xCA, 0x78, 0x00, 0x00};

// Add extra UIDs if you want to add more songs.
//byte targetUID3[] = {0x04, 0x46, 0xB3, 0xCA, 0x78, 0x00, 0x00};
//byte targetUID4[] = {0x04, 0x46, 0xB3, 0xCA, 0x78, 0x00, 0x00};

void setup() {
  Serial.begin(115200);
  SPI.begin();
  rfid.PCD_Init();
  stepper.setSpeed(7);

  // Initialize DFPlayer Mini
  mySerial.begin(9600, SERIAL_8N1, DFPLAYER_RX, DFPLAYER_TX);
  if (!myDFPlayer.begin(mySerial)) {
    Serial.println("Unable to begin:");
    Serial.println("1.Please recheck the connection!");
    Serial.println("2.Please insert the SD card!");
    while (true); // Stop execution if DFPlayer fails to initialize
  }
  Serial.println("DFPlayer Mini online.");
  myDFPlayer.volume(25);  // Set volume to 25 (0~30).
}

void loop() {
  // Check for new RFID tags
  if (rfid.PICC_IsNewCardPresent() && rfid.PICC_ReadCardSerial()) {
    lastRFIDTime = millis();

    // Print the UID of the detected card
    Serial.print("Card UID: ");
    for (byte i = 0; i < rfid.uid.size; i++) {
      Serial.print(rfid.uid.uidByte[i] < 0x10 ? " 0" : " ");
      Serial.print(rfid.uid.uidByte[i], HEX);
    }
    Serial.println();

    // Case to compare d
    if (!compareUID(rfid.uid.uidByte, lastUID, rfid.uid.size)) {
      handleNewUID(rfid.uid.uidByte, rfid.uid.size);
    } else {
      handleSameUID();
    }

    copyUID(lastUID, rfid.uid.uidByte, rfid.uid.size);
    lastUIDSize = rfid.uid.size;
  }

  // Stop the song if no tag is detected for 15 seconds.
  if (millis() - lastRFIDTime > 15000) {
    handleTimeout();
  }

  // Rotate the stepper motor if a song is playing
  if (songPlaying) {
    stepper.step(10); 
    delay(1);
  }
}

bool compareUID(byte *uid1, byte *uid2, byte length) {
  if (length != 7) {
    return false;
  }
  for (byte i = 0; i < length; i++) {
    if (uid1[i] != uid2[i]) {
      return false;
    }
  }
  return true;
}

// Compare the tags with our predefined UIDs. Add additional else-if statements to play more songs.
void handleNewUID(byte *uid, byte length) {
  stopCurrentSong();

  if (compareUID(uid, targetUID1, length)) {
    Serial.println("Playing song 1");
    myDFPlayer.play(1); // Play song 1 if UID 1 is detected
    songPlaying = true;
  } else if (compareUID(uid, targetUID2, length)) {
    Serial.println("Playing song 2");
    myDFPlayer.play(2); // Play song 2 if UID 2 is detected
    songPlaying = true;
  } 
  
  /*
  // Uncomment and add more else-if statements for additional songs
  else if (compareUID(uid, targetUID3, length)) {
    Serial.println("Playing song 3");
    myDFPlayer.play(3); // Play song 3 if UID 3 is detected
    songPlaying = true;
  } 
  else if (compareUID(uid, targetUID4, length)) {
    Serial.println("Playing song 4");
    myDFPlayer.play(4); // Play song 4 if UID 4 is detected
    songPlaying = true;
  } 
  */
  
  else {
    Serial.println("Unknown RFID tag");
  }
}

// Resume the song if the same UID is detected
void handleSameUID() {
  if (!songPlaying) {
    Serial.println("Resuming song");
    myDFPlayer.start();
    songPlaying = true;
  }
}

void handleTimeout() {
  if (songPlaying) {
    Serial.println("Timeout: Stopping current song");
    myDFPlayer.stop();
    songPlaying = false;
  }

  lastUIDSize = 0;
  lastRFIDTime = millis();
  rfid.PCD_Init();
}

void stopCurrentSong() {
  if (songPlaying) {
    Serial.println("Stopping current song");
    myDFPlayer.stop();
    songPlaying = false;
  }
}

void copyUID(byte *destUID, byte *srcUID, byte length) {
  for (byte i = 0; i < length; i++) {
    destUID[i] = srcUID[i];
  }
}
